/*
   NTBGPS.C - main module for NT kernel-mode driver BILLGPS

   Free software Copyright (c) 1996-2002 Lew Perin

   This KMD acts as "server" for the billgp "protocol" allowing the "client"
   billgPC to reach around MS/TCP to transact a bootp query/response.

   The basic purpose of this driver is to trap an incoming bootp
   reply and pass it to the caller app.  This is necessary because
   billgPC, which uses WinSock to transmit its bootp request, would
   never see the incoming reply if it relied on WinSock for input in
   a situation where the client and server disagreed on the client's
   IP address *and* the server didn't comply with the broadcast flag
   handling called out in RFC 1542.

   The original design of this driver was indebted to Russinovich and
   Cogswell's VCMON.  In that version we hooked NdisIndicateReceive
   rather than presenting ourselves as an NDIS transport.
   When we realized that not every netcard driver calls
   NdisIndicateReceive (some take the shortcut of vectoring to the
   NDIS binding context's receive handler) we did a radical redesign;
   now it's more of an orthodox NDIS transport.  The armature for the
   redesigned driver was Thomas F. Divine's transport driver skeleton
   TRANS31, which is distributed with VtoolsD.  The design for this
   driver owes much to conversations with Thomas F. Divine and P. Singh,
   as well as code Singh was nice enough to show me.

   Later, though, the service hooking tactic returned as a supplement.
   We now hook NdisSend so we can snag the outgoing bootp request and
   zero the IP source address and adjust the checksums correspondingly.
   The reason for this is to use the bootp exemption for IP stacks'
   "Martian address filtering" as discussed in RFC 1542's section 4.1.
   This became necessary when it transpired that if billgPC and a
   Novell file server's IP stack disagreed on the client's subnet the
   bootpd on the file server would never see the billgPC request.
   Note that all DOS/Windows 3.x bootp clients zero the IP source address
   in a request, so this shouldn't cause any trouble.

   As of 1.23, we're past the main disadvantage of NDIS protocolhood:
   we're now dynamically loaded rather than being loaded by NDIS when
   Windows starts up.
*/

/*
   Revision History:

   Version  Date      Reason
   -------  --------  ---------------------------------------------------
   1.27     11/24/97  Forked from version 1.26 of the BILLGPS Win95 VxD.
   1.28     11/28/97  If we call IoMarkIrpPending() in a dispatch routine
                      we return NDIS_STATUS_PENDING.
   1.29     12/14/97  We found out a bit more about token ring.
   1.30     12/21/97  We found out a bit more about asynchronous I/O.
   1.31     02/07/98  Fixed a bug in the synchronous completion of the
                      MAC address request.
   1.32     07/13/98  A bit of dead code removal.
   1.33     11/29/02  Now use Dan Lanciani's technique for NdisFreePacket
                      correctness with NT4/2000/XP.
                      We implement NDIS 4.0 for 2000/XP compatibility but
                      our bind/unbind adapters do nothing and return bad
                      status.
*/

/*
  A note on variable names: It won't take much time reading this code
  to see that we're utterly uninterested in the Microsoft variable
  naming conventions that are so widely seen in 95 and NT driver code.
  We find it usually unnecessary and distracting to be constantly told
  that e.g. a certain variable is a constant ULONG.  One stylistic
  convention we try to use religiously may need pointing out: we
  *always* give function arguments names starting with "a".  We find
  it extremely useful to know at a glance that a variable is or is not
  passed to the function by its caller.  (Thanks, Jim Hollenberg!)
*/

#pragma warning(3:4700)         // Local used w/o being initialized
#pragma warning(3:4701)         // local may be used w/o init

#define DBG 1
#include  "ntbgps.h"

#define IPPROTO_UDP 17          /* user datagram protocol */

enum {
  NetOrderPortBootpServer = 0x4300,
  NetOrderPortBootpClient = 0x4400
};

/*
   Locked Data
*/

ProtocolBlock billgpsContext;

#if 1
ULONG DbgTraceLevel = DBG_TRACE;
ULONG DbgTraceMask = DBG_RX_ALL;
#else
ULONG DbgTraceLevel = 0;
ULONG DbgTraceMask = 0;
#endif

ULONG osMajor;

/*
  Thanks to Dan Lanciani, here is a way around the facts that for
  the NT4 DDK NdisFreePacket
  - the macro version doesn't free a packet in Win2K/XP;
  - the wrapper version doesn't use the spinlock in NT, causing trouble
    in SMP
*/

VOID myNdisFreePacket(IN PNDIS_PACKET Packet)
{
  if(osMajor <= 4) {
    NdisFreePacket(Packet);
  }
  else {
#undef NdisFreePacket
#pragma warning(disable: 4210)
    VOID NdisFreePacket(IN PNDIS_PACKET Packet);
#pragma warning(default: 4210)
    NdisFreePacket(Packet);
  }
}

#define NdisFreePacket myNdisFreePacket

/* Initialized Unicode strings for which we disable indirection warning: */

#pragma warning(disable: 4057)
NDIS_STRING bgpsName = NDIS_STRING_CONST("billgPS");
NDIS_STRING devicePrefix = NDIS_STRING_CONST("\\Device\\");
NDIS_STRING symbolicLinkPrefix = NDIS_STRING_CONST("\\DosDevices\\");
NDIS_STRING tcpLinkageKeyName =
NDIS_STRING_CONST("\\Registry\\Machine\\System"
                  L"\\CurrentControlSet\\Services\\Tcpip\\Linkage");
NDIS_STRING bindValueName = NDIS_STRING_CONST("Bind");
#pragma warning(default: 4057)

/*
  The following makes building the IP and UDP headers of an outgoing
  message easier: we just copy it at open time and fill in the fields
  that can't be specified at open time: IP totalLength and checksum,
  UDP udpLength.  We also have stencils for Ethernet and Token Ring
  MAC headers.  Note that in units of greater than a byte we use Intel
  byte ordering.
*/

TxIpUdpHeader txIpUdpHeaderStencil = {
  0x45, 0, 0, 0, 0, 60, 0x11, 0, 0, 0xffffffff, // IP
  0x4400, 0x4300, 0, 0
};

RFC894Header ethernetHeaderStencil = {
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0, 0, 0, 0, 0, 0,
  PacketTypeIP
};

TokenRingHeader tokenRingHeaderStencil = {
  0x10,                         // data frame
  0x40,                         // LLC frame
  0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
  0, 0, 0, 0, 0, 0,
  0xc2, 0x70,                   // single-route broadcast, no more RI data
  0xaa, 0xaa, 3, 0, 0, 0,
  PacketTypeIP
};

/*
  Note: We once tested the interaction between compiler, linker, etc. in
  getting static data right by changing NdisMedium802_3 to NdisMediumLocalTalk
  in mediumArray; it caused the havoc we expected!
*/

enum { OurMediaMax = 2 };
NDIS_MEDIUM mediumArray[OurMediaMax] = { // shouldn't be on the stack!
  NdisMedium802_3, NdisMedium802_5
};

/*
   Locked Code
*/

/*
  Get a request entry out of the list and fill in the IRP pointer, if any.
*/

PBgpsReqNtry allocateReqNtry(POpenInstance aoiP, PIRP airpP)
{
  PBgpsReqNtry result = (PBgpsReqNtry)
    NdisInterlockedRemoveHeadList(&aoiP->requestList, &aoiP->requestSpinLock);
  if (result) result->irpP = airpP;
  else BgpsTrace("Ran out of request entries");
  return result;
}

BOOLEAN prepareMACAddressRequest(POpenInstance aoiP, PNDIS_REQUEST arequestP)
{
  switch(aoiP->u1.medium)
    {
    case NdisMedium802_3:
      arequestP->DATA.QUERY_INFORMATION.Oid = OID_802_3_CURRENT_ADDRESS;
      break;
    case NdisMedium802_5:
      arequestP->DATA.QUERY_INFORMATION.Oid = OID_802_5_CURRENT_ADDRESS;
      break;
    default:
      BgpsTrace2("MAC address request for unknown medium: %x",
                 aoiP->u1.medium);
      return FALSE;
    }
  arequestP->RequestType = NdisRequestQueryInformation;
  arequestP->DATA.QUERY_INFORMATION.InformationBuffer = aoiP->macAddress;
  arequestP->DATA.QUERY_INFORMATION.InformationBufferLength = MacAddressSize;
  return TRUE;
}

/*
  We need to set the filter on this adapter.
*/

void prepareSetFilterRequest(POpenInstance aoiP, PNDIS_REQUEST arequestP,
                             UINT afilter)
{
  aoiP->packetFilter = afilter;
  arequestP->RequestType = NdisRequestSetInformation;
  arequestP->DATA.SET_INFORMATION.Oid = OID_GEN_CURRENT_PACKET_FILTER;
  arequestP->DATA.SET_INFORMATION.InformationBuffer = &aoiP->packetFilter;
  arequestP->DATA.SET_INFORMATION.InformationBufferLength =
    sizeof(aoiP->packetFilter);
}

/*
  Just a convenience function that'll do all we need to do to finish off
  an IRP in one line of code and return the status.
*/

NTSTATUS finishIrp(IN OUT PIRP airpP, IN NTSTATUS astatus, IN ULONG aninfo)
{
  airpP->IoStatus.Status = astatus;
  airpP->IoStatus.Information = aninfo;
  IoCompleteRequest(airpP, IO_NO_INCREMENT);
  return astatus;
}

/*
   Entry point for Ioctl requests. The following functions are supported:

   DIOC_BillgpAwaitBootp: App is ready for a bootp packet.

   DIOC_BillgpAwaitAny: App is ready for a packet, any packet!

   DIOC_BillgpGetVersion: What version of the driver is this?

   DIOC_BillgpGetMacAddress: MAC address of first (only?) netcard, please!

   DIOC_BillgpGetHWType: bootp (not NDIS) hardware type of MAC.

   DIOC_BillgpDumpDriver: Blow the driver's data into the app's buffer.
*/

NTSTATUS BgpsIoControl(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  NTSTATUS status = STATUS_SUCCESS;
  UINT lengthNeeded;            // how much buffer space we need to write to
  UINT lengthTransferred = 0;   // how much buffer space we did write to

  BgpsTrace2("Ioctl function = %x",
             ioslP->Parameters.DeviceIoControl.IoControlCode);

  switch (ioslP->Parameters.DeviceIoControl.IoControlCode) {

  case DIOC_BillgpAwaitBootp:
  case DIOC_BillgpAwaitAny:
    return finishIrp(airpP, NDIS_STATUS_FAILURE, 0);

  case DIOC_BillgpGetVersion:
    lengthNeeded = sizeof(UINT);
    if ((airpP->AssociatedIrp.SystemBuffer == NULL) ||
        (ioslP->Parameters.DeviceIoControl.OutputBufferLength <
         lengthNeeded)) {
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      *((UINT*)airpP->AssociatedIrp.SystemBuffer) =
        BILLGPS_Major * 100 + BILLGPS_Minor;
      lengthTransferred = lengthNeeded;
    }
    return finishIrp(airpP, status, lengthTransferred);

  case DIOC_BillgpGetMacAddress:
    lengthNeeded = MacAddressSize;
    if ((airpP->AssociatedIrp.SystemBuffer == NULL) ||
        (ioslP->Parameters.DeviceIoControl.OutputBufferLength <
         lengthNeeded)) {
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      RtlCopyBytes(airpP->AssociatedIrp.SystemBuffer, oiP->macAddress,
                   MacAddressSize);
      lengthTransferred = lengthNeeded;
    }
    return finishIrp(airpP, status, lengthTransferred);

  case DIOC_BillgpGetHWType:
    lengthNeeded = sizeof(UCHAR);
    if ((airpP->AssociatedIrp.SystemBuffer == NULL) ||
        (ioslP->Parameters.DeviceIoControl.OutputBufferLength <
         lengthNeeded)) {
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      switch((NDIS_MEDIUM)(oiP->u1.medium)) {
      case NdisMedium802_3:
        *((UCHAR*)airpP->AssociatedIrp.SystemBuffer) =
          (UCHAR)BootpHWTypeEthernet;
        lengthTransferred = lengthNeeded;
        break;
      case NdisMedium802_5:
        *((UCHAR*)airpP->AssociatedIrp.SystemBuffer) = (UCHAR)BootpHWTypeIEEE;
        lengthTransferred = lengthNeeded;
        break;
      default:
        status = NDIS_STATUS_UNSUPPORTED_MEDIA;
      }
    }
    return finishIrp(airpP, status, lengthTransferred);

  case DIOC_BillgpDumpDriver:
    lengthNeeded = sizeof(OpenInstance);
    if ((airpP->AssociatedIrp.SystemBuffer == NULL) ||
        (ioslP->Parameters.DeviceIoControl.OutputBufferLength <
         lengthNeeded)) {
      status = STATUS_INVALID_PARAMETER;
    }
    else {
      RtlCopyBytes(airpP->AssociatedIrp.SystemBuffer, oiP,
                   lengthNeeded);
      lengthTransferred = lengthNeeded;
    }
    return finishIrp(airpP, status, lengthTransferred);

  default:
    return finishIrp(airpP, NDIS_STATUS_FAILURE, 0);
  }

}

/*
  It's the MACs that bind to TCP/IP that we want to bind to.  We find them
  here in the form of a dynamically allocated KEY_VALUE_PARTIAL_INFORMATION
  whose Data member has what we need in the form of a REG_MULTI_SZ array.
  We query the appropriate key twice, the first time just to ascertain the
  size of what we'll allocate and the second time "for real".  If anything
  goes wrong we return NULL.  It's the caller's responsibility to free
  what we return.
*/

PKEY_VALUE_PARTIAL_INFORMATION getTcpBindings(void)
{
  PKEY_VALUE_PARTIAL_INFORMATION result = NULL; // what we return 
  OBJECT_ATTRIBUTES objAttrs;
  NTSTATUS status;
  HANDLE keyHandle;

  InitializeObjectAttributes(&objAttrs, &tcpLinkageKeyName,
                             OBJ_CASE_INSENSITIVE, NULL, NULL);
  status = ZwOpenKey(&keyHandle, KEY_READ, &objAttrs);
  if (!NT_SUCCESS(status)) {
    BgpsTrace3("Status of %x opening %ws", status, tcpLinkageKeyName.Buffer);
  }
  else {
    ULONG resultLength;
    KEY_VALUE_PARTIAL_INFORMATION valueInfo;

    BgpsTrace2("Opened %ws", tcpLinkageKeyName.Buffer);

    status = ZwQueryValueKey(keyHandle, &bindValueName,
                             KeyValuePartialInformation, &valueInfo,
                             sizeof(valueInfo), &resultLength);
    if (!NT_SUCCESS(status) && (status != STATUS_BUFFER_OVERFLOW)) {
      BgpsTrace2("Status of %x querying key value for size", status);
    }
    else {                      // We know how big it needs to be.
      ULONG valueInfoLength = valueInfo.DataLength +
        FIELD_OFFSET(KEY_VALUE_PARTIAL_INFORMATION, Data[0]);
      PKEY_VALUE_PARTIAL_INFORMATION valueInfoP =
        (PKEY_VALUE_PARTIAL_INFORMATION)
        ExAllocatePool(PagedPool, valueInfoLength);
      if (valueInfoP != NULL) {
        status = ZwQueryValueKey(keyHandle, &bindValueName,
                                 KeyValuePartialInformation,
                                 valueInfoP,
                                 valueInfoLength, &resultLength);
        if (!NT_SUCCESS(status)) {
          BgpsTrace2("Status of %x querying key value", status);
        }
        else if (valueInfoLength != resultLength) {
          BgpsTrace3("Querying key value result len = %u "
                     "but previous len = %u",
                     resultLength, valueInfoLength);
        }
        else if (valueInfoP->Type != REG_MULTI_SZ) {
          BgpsTrace2("Tcpip bind value not REG_MULTI_SZ but %u",
                     valueInfoP->Type);
        }
        else {                  // It's OK.
#if DBG
          ULONG i;
          WCHAR* dataP = (WCHAR*)(&valueInfoP->Data[0]);
          BgpsTrace("Bind value:");
          for (i = 0; *dataP != UNICODE_NULL; i++) {
            UNICODE_STRING macName;
            RtlInitUnicodeString(&macName, dataP);
            BgpsTrace3("Mac %u = %ws", i, macName.Buffer);
            dataP +=
              (macName.Length + sizeof(UNICODE_NULL)) / sizeof(WCHAR);
          }
#endif // DBG
          result = valueInfoP;
        }
      }
    }
    ZwClose(keyHandle);
  }
  return result;
}

NTSTATUS registerProtocol(PNDIS_HANDLE andisProtocolHandleP)
{
  NDIS40_PROTOCOL_CHARACTERISTICS npc;
  NDIS_STATUS result;

  /*
    Initialize the PROTOCOL characteristics for the call to
    NdisRegisterProtocol.
  */

  NdisZeroMemory(&npc, sizeof(npc));
  npc.MajorNdisVersion = BILLGPS_NDIS_MAJOR_VERSION;
  npc.MinorNdisVersion = BILLGPS_NDIS_MINOR_VERSION;
  npc.Reserved = 0;
  npc.OpenAdapterCompleteHandler = BgpsOpenAdapterCompleteHandler;
  npc.CloseAdapterCompleteHandler = BgpsCloseAdapterCompleteHandler;
  npc.SendCompleteHandler = BgpsSendCompleteHandler;
  npc.TransferDataCompleteHandler = BgpsTransferDataCompleteHandler;
  npc.ResetCompleteHandler = BgpsResetCompleteHandler;
  npc.RequestCompleteHandler = BgpsRequestCompleteHandler;
  npc.ReceiveHandler = BgpsReceiveHandler;
  npc.ReceiveCompleteHandler = BgpsReceiveCompleteHandler;
  npc.StatusHandler = BgpsStatusHandler;
  npc.StatusCompleteHandler = BgpsStatusCompleteHandler;
  npc.BindAdapterHandler = BgpsBindAdapterHandler;
  npc.UnbindAdapterHandler = BgpsUnbindAdapterHandler;
  npc.Name = bgpsName;

  BgpsTrace("Calling NdisRegisterProtocol...");

  NdisRegisterProtocol(&result, andisProtocolHandleP,
                       (PNDIS_PROTOCOL_CHARACTERISTICS)&npc, sizeof(npc));

  if(result != NDIS_STATUS_SUCCESS) {
    BgpsTrace2("NdisRegisterProtocol failed with code 0x%X\n", result);
  }
  else {
    BgpsTrace("NdisRegisterProtocol succeeded");
  }

  return result;
}

void fillDriverObject(IN PDRIVER_OBJECT adriverObjectP)
{
  adriverObjectP->MajorFunction[IRP_MJ_CREATE] = BgpsOpen;
  adriverObjectP->MajorFunction[IRP_MJ_CLOSE] = BgpsClose;
  adriverObjectP->MajorFunction[IRP_MJ_READ] = BgpsRead;
  adriverObjectP->MajorFunction[IRP_MJ_WRITE] = BgpsWrite;
  adriverObjectP->MajorFunction[IRP_MJ_CLEANUP] = BgpsCleanup;
  adriverObjectP->MajorFunction[IRP_MJ_DEVICE_CONTROL] = BgpsIoControl;
  adriverObjectP->DriverUnload = BgpsUnload;
}

/*
  Before we fill in the device extension we make sure it's zeroed so the
  matching "empty" function can be sure any non-NULL dynamically
  allocated pointer should be freed.

  If the MAC name is \Device\Elnk31, the device name will be
  \Device\billgPSElnk31 and the symbolic link billgPSElnk31.
*/

BOOLEAN fillDeviceExtension(IN OUT PDeviceExtension adevExtP,
                            IN PDEVICE_OBJECT adevObjP,
                            IN PUNICODE_STRING amacNameP,
                            IN PUNICODE_STRING adeviceNameP)
{
  NTSTATUS status;

  NdisZeroMemory(adevExtP, sizeof(DeviceExtension));
  adevExtP->deviceObjectP = adevObjP;
  adevExtP->macName.MaximumLength =
    (USHORT)(amacNameP->Length + sizeof(UNICODE_NULL));
  adevExtP->macName.Buffer = ExAllocatePool(NonPagedPool,
                                            adevExtP->macName.MaximumLength);
  if (adevExtP->macName.Buffer == NULL) return FALSE;
  RtlAppendUnicodeStringToString(&adevExtP->macName, amacNameP);
  adevExtP->deviceName.MaximumLength =
    (USHORT)(adeviceNameP->Length + sizeof(UNICODE_NULL));
  adevExtP->deviceName.Buffer =
    ExAllocatePool(NonPagedPool, adevExtP->deviceName.MaximumLength);
  if (adevExtP->deviceName.Buffer == NULL) return FALSE;
  RtlCopyUnicodeString(&adevExtP->deviceName, adeviceNameP);
  BgpsTrace2("Device name in device extension: %ws",
             adevExtP->deviceName.Buffer);
  adevExtP->symbolicLink.MaximumLength =
    (USHORT)(symbolicLinkPrefix.Length + adeviceNameP->Length
             - devicePrefix.Length + sizeof(UNICODE_NULL));
  adevExtP->symbolicLink.Buffer =
    ExAllocatePool(NonPagedPool, adevExtP->symbolicLink.MaximumLength);
  if (adevExtP->symbolicLink.Buffer == NULL) return FALSE;
  RtlAppendUnicodeStringToString(&adevExtP->symbolicLink, &symbolicLinkPrefix);
  RtlAppendUnicodeToString(&adevExtP->symbolicLink,
                           adevExtP->deviceName.Buffer +
                           devicePrefix.Length / sizeof(WCHAR));
  BgpsTrace2("Symbolic link: %ws", adevExtP->symbolicLink.Buffer);
  status = IoCreateSymbolicLink(&adevExtP->symbolicLink, adeviceNameP);
  if (!NT_SUCCESS(status)) {
    BgpsTrace2("IoCreateSymbolicLink status = %x", status);
    RtlInitUnicodeString(&adevExtP->symbolicLink, NULL);
    // return FALSE;
  }
  return TRUE;
}

/*
  This convenience function deallocates and zeroes the pointer pointed to
  by the argument only if it isn't already NULL.
*/

void empty(PVOID* memPP)
{
  if (*memPP != NULL) {
    ExFreePool(*memPP);
    *memPP = NULL;
  }
}

void emptyDeviceExtension(IN PDeviceExtension adevExtP)
{
  if (adevExtP->symbolicLink.Buffer != NULL) {
    NTSTATUS status = IoDeleteSymbolicLink(&adevExtP->symbolicLink);
    if (!NT_SUCCESS(status)) {
      BgpsTrace2("IoDeleteSymbolicLink status = %x", status);
    }
    empty(&adevExtP->symbolicLink.Buffer);
  }
  empty(&adevExtP->macName.Buffer);
  empty(&adevExtP->deviceName.Buffer);
}

/*
  Here we take down everything associated with a device object, including
  what goes with its extension.
*/

void deleteDevice(IN OUT PDEVICE_OBJECT adevObjP)
{
  PDeviceExtension devExtP = (PDeviceExtension)adevObjP->DeviceExtension;
  if (devExtP) emptyDeviceExtension(devExtP);
  IoDeleteDevice(adevObjP);
}

BOOLEAN createDevice(IN OUT PDRIVER_OBJECT adriverObjectP,
                     IN PUNICODE_STRING amacNameP)
{
  NTSTATUS status;
  PDEVICE_OBJECT devObjP;
  UNICODE_STRING deviceName;
  BOOLEAN result = FALSE;

  BgpsTrace2("createDevice for MAC %ws", amacNameP->Buffer);
  if (RtlCompareMemory(amacNameP->Buffer, devicePrefix.Buffer,
                       devicePrefix.Length) < devicePrefix.Length) {
    return result;
  }

  deviceName.Length = 0;
  deviceName.MaximumLength =
    (USHORT)(amacNameP->Length + bgpsName.Length + sizeof(UNICODE_NULL));
  deviceName.Buffer = ExAllocatePool(PagedPool, deviceName.MaximumLength);
  if (deviceName.Buffer != NULL) {
    RtlAppendUnicodeStringToString(&deviceName, &devicePrefix);
    RtlAppendUnicodeStringToString(&deviceName, &bgpsName);
    RtlAppendUnicodeToString(&deviceName, amacNameP->Buffer +
                             devicePrefix.Length / sizeof(WCHAR));
    BgpsTrace2("Device name: %ws", deviceName.Buffer);
    status = IoCreateDevice(adriverObjectP, sizeof(DeviceExtension),
                            &deviceName, FILE_DEVICE_TRANSPORT, 0, FALSE,
                            &devObjP);
    if (NT_SUCCESS(status)) {
      PDeviceExtension devExtP = (PDeviceExtension)devObjP->DeviceExtension;
      if (fillDeviceExtension(devExtP, devObjP, amacNameP, &deviceName)){
        devObjP->Flags |= DO_DIRECT_IO;
        result = TRUE;
      }
      else {
        BgpsTrace("Couldn't fill device extension");
        deleteDevice(devObjP);
      }
    }
    else BgpsTrace2("IoCreateDevice status = %x", status);
  }
  if (deviceName.Length > 0) ExFreePool(deviceName.Buffer);
  return result;
}

/*
   Main entry point for NDIS driver; called during initialization

   The registry path argument is bogus for a dynamic load, so we ignore it.
   Instead, we find the lower-edge bindings for TCP/IP in the Registry and
   try to create a transport protocol device for each one.
*/

NTSTATUS DriverEntry(IN PDRIVER_OBJECT adriverObjectP,
                     IN PUNICODE_STRING aregistryPathP)
{
  NTSTATUS result = NDIS_STATUS_FAILURE;
  PKEY_VALUE_PARTIAL_INFORMATION tcpBindingsP;
  ULONG osMinor;

  BgpsTrace4("%ws: version %d.%d DriverEntry",
             bgpsName.Buffer, BILLGPS_Major, BILLGPS_Minor);

  PsGetVersion(&osMajor, &osMinor, NULL, NULL);
  BgpsTrace3("NT version %d.%d\n", osMajor, osMinor);

  tcpBindingsP = getTcpBindings();
  if (tcpBindingsP != NULL) {
    result = registerProtocol(&billgpsContext.ndisProtocolHandle);
    if (result == NDIS_STATUS_SUCCESS) {
      WCHAR* bindP = (WCHAR*)(&tcpBindingsP->Data[0]);
      UNICODE_STRING macName;
      fillDriverObject(adriverObjectP);
      for (; *bindP != UNICODE_NULL;
           bindP += (macName.Length + sizeof(UNICODE_NULL)) / sizeof(WCHAR)) {
        RtlInitUnicodeString(&macName, bindP);
        createDevice(adriverObjectP, &macName);
      }
      if (adriverObjectP->DeviceObject == NULL) {
        BgpsTrace("No devices created");
        result = STATUS_UNSUCCESSFUL;
      }
    }
    else billgpsContext.ndisProtocolHandle = 0;
    ExFreePool(tcpBindingsP);
  }
  return result;   
}

VOID BgpsUnload(IN PDRIVER_OBJECT adriverObjectP)
{
  NDIS_STATUS status;
  PDEVICE_OBJECT devObjP;
  PDEVICE_OBJECT nextDevObjP;

  BgpsTrace("BgpsUnload");
  for (devObjP = (PDEVICE_OBJECT)adriverObjectP->DeviceObject;
       devObjP != NULL;
       devObjP = nextDevObjP) {
    nextDevObjP = devObjP->NextDevice;
    deleteDevice(devObjP);
  }

  if (billgpsContext.ndisProtocolHandle != 0) {
    NdisDeregisterProtocol(&status, billgpsContext.ndisProtocolHandle);
  }
  BgpsTrace2("NdisDeregisterProtocol status = %x", status);
}

/*
  This functionette makes an IRP ready to be submitted to an NDIS wrapper
  function that may complete asynchronously.
*/

void pendIrp(IN OUT PIRP airpP)
{
  IoMarkIrpPending(airpP);
  airpP->IoStatus.Status = STATUS_PENDING;
}

/*
  Here we allocate and populate an OpenInstance, returning a status indicating
  whether we succeeded.  Note that even if we fail we might have allocated
  and partially filled the OpenInstance.
*/

NDIS_STATUS fillOpenInstance(POpenInstance* aoiPP,
                             PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  NDIS_STATUS result = NDIS_STATUS_SUCCESS;
  int i;
  POpenInstance oiP =
     (POpenInstance)ExAllocatePool(NonPagedPool, sizeof(OpenInstance));
  *aoiPP = oiP;
  if (!oiP) return STATUS_INSUFFICIENT_RESOURCES;
  NdisZeroMemory(oiP, sizeof(OpenInstance));
  oiP->deviceObjectP = adeviceObjectP;
  oiP->openOrCloseIrpP = airpP;
  InitializeListHead(&oiP->rcvList); // guarded by Cancel Spinlock
  BgpsTrace2("Address of rcvList = %x", &oiP->rcvList);
  NdisAllocateSpinLock(&oiP->requestSpinLock);
  InitializeListHead(&oiP->requestList);
  for (i = 0; i < MAX_REQUESTS; i++) {
    NdisInterlockedInsertTailList(&oiP->requestList,
                                  &oiP->requests[i].listElement,
                                  &oiP->requestSpinLock);

  }
  NdisAllocateSpinLock(&oiP->txMacHeaderSpinLock);
  InitializeListHead(&oiP->txMacHeaderList);
  for (i = 0; i < MAX_TRANSMITS; i++) {
    NdisInterlockedInsertTailList(&oiP->txMacHeaderList,
                                  &oiP->txMacHeaders[i].listElement,
                                  &oiP->txMacHeaderSpinLock);
  }
  NdisAllocateSpinLock(&oiP->txIpUdpHeaderSpinLock);
  InitializeListHead(&oiP->txIpUdpHeaderList);
  for (i = 0; i < MAX_TRANSMITS; i++) {
    NdisInterlockedInsertTailList(&oiP->txIpUdpHeaderList,
                                  &oiP->txIpUdpHeaders[i].listElement,
                                  &oiP->txIpUdpHeaderSpinLock);
    RtlCopyBytes(&oiP->txIpUdpHeaders[i].payload, &txIpUdpHeaderStencil,
                 sizeof(TxIpUdpHeader));
      

  }
  NdisAllocatePacketPool(&result, &(oiP->packetPool), NPackets,
                         sizeof(BgpsReserved));
  if (result != NDIS_STATUS_SUCCESS) {
    BgpsTrace("Couldn't allocate packet pool");
    oiP->packetPool = NULL;
  }
  return result;
}

void emptyOpenInstance(POpenInstance aoiP)
{
  NdisFreeSpinLock(&aoiP->requestSpinLock);
  NdisFreeSpinLock(&aoiP->txMacHeaderSpinLock);
  NdisFreeSpinLock(&aoiP->txIpUdpHeaderSpinLock);
  if (aoiP->packetPool) {
    NdisFreePacketPool(aoiP->packetPool);
    aoiP->packetPool = NULL;
  }
}

NTSTATUS BgpsOpen(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PDeviceExtension deP = (PDeviceExtension)adeviceObjectP->DeviceExtension;
  POpenInstance oiP;
  NDIS_STATUS status = fillOpenInstance(&oiP, adeviceObjectP, airpP);
  BgpsTrace("BgpsOpen");
  if (NT_SUCCESS(status)) {
    PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
    ioslP->FileObject->FsContext = oiP;
    pendIrp(airpP);
    NdisOpenAdapter(&status, &oiP->openErrorStatus, &oiP->adapterHandle,
                    &oiP->u1.selectedMediumIndex, mediumArray, OurMediaMax,
                    billgpsContext.ndisProtocolHandle, (NDIS_HANDLE)oiP,
                    &deP->macName, 0, NULL);
    if (status != NDIS_STATUS_PENDING) {
      BgpsOpenAdapterCompleteHandler((NDIS_HANDLE)oiP, status,
                                     oiP->openErrorStatus);
    }
    else BgpsTrace("NdisOpenAdapter pending");
    return NDIS_STATUS_PENDING;
  }
  else {                        // ran out of resources
    if (oiP){
      emptyOpenInstance(oiP); // ...after allocating OpenInstance
      ExFreePool(oiP);
    }
    return finishIrp(airpP, status, 0);
  }
}

NTSTATUS BgpsClose(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  NDIS_STATUS status = NDIS_STATUS_SUCCESS;

  BgpsTrace("BgpsClose");
  oiP->openOrCloseIrpP = airpP;
  pendIrp(airpP);
  NdisCloseAdapter(&status, oiP->adapterHandle);
  if (status != NDIS_STATUS_PENDING) {
    BgpsCloseAdapterCompleteHandler((NDIS_HANDLE)oiP, status);
  }
  else BgpsTrace("NdisCloseAdapter pending");
#if 0
  if (status == NDIS_STATUS_SUCCESS) return NDIS_STATUS_PENDING;
  else return status;
#else
  return NDIS_STATUS_PENDING;
#endif
}

/*
  This function, which computes the checksum of an IP header, was adapted
  from cksum() in Comer & Stevens, Internetworking with TCP/IP, v.2,
  section 5.4.4.
*/

void computeIPChecksum(IPHeader* aipHeaderP)
{
  ULONG sum;
  USHORT* wordP = (USHORT*)aipHeaderP;
  UINT nWords = (aipHeaderP->versionAndHeaderLength & 0xf) * 2;
  aipHeaderP->checksum = 0;    /* Initialize embedded checksum. */
  for (sum = 0; nWords > 0; nWords--, wordP++) {
    sum += htons(*wordP);
  }
  sum = (sum >> 16) + (sum & 0xffff); /* Add in carry. */
  sum += (sum >> 16);           /* Maybe one more. */
  sum = ~sum;
  aipHeaderP->checksum = htons(sum);
}

NTSTATUS BgpsWrite(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  NDIS_STATUS status;
  PNDIS_PACKET packetP;
  PTxMacHeaderNtry macHeaderNtryP;
  PTxIpUdpHeaderNtry ipUdpHeaderNtryP;

  BgpsTrace3("BgpsWrite w/ write length %u, MDL length %u",
             ioslP->Parameters.Write.Length,
             MmGetMdlByteCount(airpP->MdlAddress));

  NdisAllocatePacket(&status, &packetP, oiP->packetPool);
  if (status != NDIS_STATUS_SUCCESS) {
    BgpsTrace2("Packet allocation failure on write: %x", status);
    return finishIrp(airpP, status, 0);
  }

  RESERVED(packetP)->irpP = airpP;
  NdisChainBufferAtFront(packetP, airpP->MdlAddress);

  ipUdpHeaderNtryP = (PTxIpUdpHeaderNtry)
    NdisInterlockedRemoveHeadList(&oiP->txIpUdpHeaderList,
                                  &oiP->txIpUdpHeaderSpinLock);
  if (ipUdpHeaderNtryP) {
    PMDL ipUdpHeaderMdlP =
      IoAllocateMdl(&ipUdpHeaderNtryP->payload,
                    sizeof(IPHeader) + sizeof(UDPHeader), FALSE, FALSE, NULL);
      if (ipUdpHeaderMdlP != NULL) {
        UINT udpLength = MmGetMdlByteCount(airpP->MdlAddress) +
          sizeof(UDPHeader);
        MmBuildMdlForNonPagedPool(ipUdpHeaderMdlP);
        ipUdpHeaderNtryP->payload.udp.udpLength = htons(udpLength);
        ipUdpHeaderNtryP->payload.ip.totalLength =
          htons(udpLength + sizeof(IPHeader));
        computeIPChecksum(&ipUdpHeaderNtryP->payload.ip);
        NdisChainBufferAtFront(packetP, ipUdpHeaderMdlP);

        macHeaderNtryP = (PTxMacHeaderNtry)
          NdisInterlockedRemoveHeadList(&oiP->txMacHeaderList,
                                        &oiP->txMacHeaderSpinLock);
        if (macHeaderNtryP) {
          ULONG macHeaderLength = (oiP->u1.medium == NdisMedium802_3) ?
            sizeof(RFC894Header) : sizeof(TokenRingHeader);
          PMDL macHeaderMdlP =
            IoAllocateMdl(&macHeaderNtryP->payload, macHeaderLength,
                          FALSE, FALSE, NULL);
          if (macHeaderMdlP != NULL) {
            MmBuildMdlForNonPagedPool(macHeaderMdlP);
            NdisChainBufferAtFront(packetP, macHeaderMdlP);
            pendIrp(airpP);
            NdisSend(&status, oiP->adapterHandle, packetP);
            BgpsTrace2Mask(DBG_SEND, "NdisSend status = %x", status);
          }
          else {
            BgpsTrace("Couldn't allocate TX MAC header MDL");
            NdisInterlockedInsertTailList(&oiP->txMacHeaderList,
                                          &macHeaderNtryP->listElement,
                                          &oiP->txMacHeaderSpinLock);
          }
        }
        else {
          BgpsTrace("Ran out of TX MAC header entries");
          status = STATUS_INSUFFICIENT_RESOURCES;
        }
      }
      else {
        BgpsTrace("Couldn't allocate TX IP/UDP header MDL");
        NdisInterlockedInsertTailList(&oiP->txIpUdpHeaderList,
                                      &ipUdpHeaderNtryP->listElement,
                                      &oiP->txIpUdpHeaderSpinLock);
      }
  }
  else {
    BgpsTrace("Ran out of TX IP/UDP header entries");
    status = STATUS_INSUFFICIENT_RESOURCES;
  }
  if (status != NDIS_STATUS_PENDING) { // Packet was allocated, at least
    BgpsSendCompleteHandler(oiP, packetP, status);
    if (status != NDIS_STATUS_SUCCESS) return status;
  }
#if 0
  if (status == NDIS_STATUS_SUCCESS) return NDIS_STATUS_PENDING;
  else return NDIS_STATUS_PENDING;
#else
  if (status == STATUS_INSUFFICIENT_RESOURCES) {
    return STATUS_INSUFFICIENT_RESOURCES;
  }
  else return NDIS_STATUS_PENDING;
#endif
}

NTSTATUS BgpsRead(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  NDIS_STATUS status;
  PNDIS_PACKET packetP;
  KIRQL irql;

  BgpsTrace3("BgpsRead w/ read length %u, MDL length %u",
             ioslP->Parameters.Read.Length,
             MmGetMdlByteCount(airpP->MdlAddress));

  NdisAllocatePacket(&status, &packetP, oiP->packetPool);
  if (status != NDIS_STATUS_SUCCESS) {
    BgpsTrace2("Packet allocation failure on read: %x", status);
    return finishIrp(airpP, status, 0);
  }

  RESERVED(packetP)->irpP = airpP;
  pendIrp(airpP);
  NdisChainBufferAtFront(packetP, airpP->MdlAddress);
  IoAcquireCancelSpinLock(&irql);
  InsertTailList(&oiP->rcvList, &RESERVED(packetP)->listElement);
  IoSetCancelRoutine(airpP, BgpsCancelReadIrp); // Now it's cancelable
  IoReleaseCancelSpinLock(irql);
  // STATUS_SUCCESS = BSOD from multiple IRP completion!
  return STATUS_PENDING;        
}

NTSTATUS BgpsCleanup(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  PLIST_ENTRY listEntryP;

  BgpsTrace("BgpsCleanup");

  pendIrp(airpP);

  do {
    KIRQL irql;
    IoAcquireCancelSpinLock(&irql);
    listEntryP = (IsListEmpty(&oiP->rcvList)) ? NULL :
      RemoveHeadList(&oiP->rcvList);
    IoReleaseCancelSpinLock(irql);
    if (listEntryP != NULL) {
      PBgpsReserved reservedP =
        CONTAINING_RECORD(listEntryP, BgpsReserved, listElement);
      PNDIS_PACKET packetP =
        CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
      BgpsTrace("Completing read in Cleanup");
      BgpsTransferDataCompleteHandler(oiP, packetP, STATUS_CANCELLED, 0);
    }
  } while (listEntryP != NULL);
  return finishIrp(airpP, STATUS_SUCCESS, 0);
}

/*
  After successfully opening the adapter, if it's a real LAN adapter we
  want to get its MAC address so the app can ask us for it later.
  We also need to set its filter so we'll see incoming packets.

  The IRP will be finished off in NdisRequestComplete processing if all
  goes well.
*/

VOID
BgpsOpenAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                               IN NDIS_STATUS astatus,
                               IN NDIS_STATUS aopenErrorStatus)
{
  OpenInstance* oiP = (OpenInstance*)abindingContext;
  PIRP irpP = oiP->openOrCloseIrpP;

  BgpsTrace3("OpenAdapterComplete status / openerror = %x / %x",
             astatus, aopenErrorStatus);

  if (astatus == NDIS_STATUS_SUCCESS) {
    PBgpsReqNtry macAddressReqNtryP = allocateReqNtry(oiP, NULL);
    PBgpsReqNtry setFilterReqNtryP = allocateReqNtry(oiP, irpP);
    if (macAddressReqNtryP && setFilterReqNtryP) {
      if (prepareMACAddressRequest(oiP, &macAddressReqNtryP->request)) {
        NdisRequest(&oiP->macAddressReqStatus, oiP->adapterHandle,
                    &macAddressReqNtryP->request);
        if (oiP->macAddressReqStatus != NDIS_STATUS_PENDING) {
          BgpsRequestCompleteHandler((NDIS_HANDLE)oiP,
                                     &macAddressReqNtryP->request,
                                     oiP->macAddressReqStatus);
          
        }
        prepareSetFilterRequest(oiP, &setFilterReqNtryP->request,
                                NDIS_PACKET_TYPE_DIRECTED |
                                NDIS_PACKET_TYPE_BROADCAST);
        NdisRequest(&oiP->setFilterReqStatus, oiP->adapterHandle,
                    &setFilterReqNtryP->request);
        if (oiP->setFilterReqStatus != NDIS_STATUS_PENDING) {
          BgpsRequestCompleteHandler((NDIS_HANDLE)oiP,
                                     &setFilterReqNtryP->request,
                                     oiP->setFilterReqStatus);
        }
      }
      else astatus = NDIS_STATUS_UNSUPPORTED_MEDIA; // shouldn't happen!
    }
    else {
      astatus = STATUS_INSUFFICIENT_RESOURCES;
    }
  }
  if (astatus != NDIS_STATUS_SUCCESS) {
    // Never launched an NdisRequest
    finishIrp(irpP, astatus, 0);
    emptyOpenInstance(oiP);
    ExFreePool(oiP);
  }
}

/*
  When NDIS tells us the unbinding of an adapter's complete, we want to
  make sure we don't keep a dangling handle for it.
*/

VOID BgpsCloseAdapterCompleteHandler(IN NDIS_HANDLE abindingContext,
                                     IN NDIS_STATUS astatus)
{
  OpenInstance* oiP = (OpenInstance*)abindingContext;
  PIRP irpP = oiP->openOrCloseIrpP;

  BgpsTrace2("CloseAdapterCompleteHandler status = %x", astatus);

  emptyOpenInstance(oiP);

  finishIrp(irpP, astatus, 0);
  ExFreePool(oiP);
}

/*
  Here, because we need to handle the freeing of MDLs we tried to create
  in the write dispatch routine, we handle all business of that type, even
  if the write dispatch routine ran into trouble before calling NdisSend.
  We have a packet to the back of which is chained the MDL for the user
  level data; there may or may not be two more MDLs in front of the user
  data.  We don't need to dispose of the user data MDL; NT is nice enough
  to do that for us.
*/

VOID BgpsSendCompleteHandler(IN NDIS_HANDLE abindingContext,
                             IN PNDIS_PACKET apacketP,
                             IN NDIS_STATUS astatus)
{
  OpenInstance* oiP = (OpenInstance*)abindingContext;
  BgpsReserved* reservedP = (BgpsReserved*)(apacketP->ProtocolReserved);
  PIRP irpP = reservedP->irpP;
  UINT userBytesSent = (astatus == NDIS_STATUS_SUCCESS) ?
    MmGetMdlByteCount(irpP->MdlAddress) : 0;
  UINT bufferCount;

  BgpsTrace2Mask(DBG_SEND, "SendCompleteHandler status = %x", astatus);
#pragma warning(disable: 4127 4702)
  NdisQueryPacket(apacketP, NULL, &bufferCount, NULL, NULL);
#pragma warning(default: 4127 4702)

  if (bufferCount > 2) {
    PNDIS_BUFFER macHeaderMdlP;
    MacHeaderPayload* macHeaderPayloadP;
    PTxMacHeaderNtry txMacHeaderNtryP;
    NdisUnchainBufferAtFront(apacketP, &macHeaderMdlP);
    macHeaderPayloadP = MmGetSystemAddressForMdl(macHeaderMdlP);
    txMacHeaderNtryP =
      CONTAINING_RECORD(macHeaderPayloadP, TxMacHeaderNtry, payload);
    NdisInterlockedInsertTailList(&oiP->txMacHeaderList,
                                  &txMacHeaderNtryP->listElement,
                                  &oiP->txMacHeaderSpinLock);
  }

  if (bufferCount > 1) {
    PNDIS_BUFFER ipUdpHeaderMdlP;
    TxIpUdpHeader* ipUdpHeaderPayloadP;
    PTxIpUdpHeaderNtry txIpUdpHeaderNtryP;
    NdisUnchainBufferAtFront(apacketP, &ipUdpHeaderMdlP);
    ipUdpHeaderPayloadP = MmGetSystemAddressForMdl(ipUdpHeaderMdlP);
    txIpUdpHeaderNtryP =
      CONTAINING_RECORD(ipUdpHeaderPayloadP, TxIpUdpHeaderNtry, payload);
    NdisInterlockedInsertTailList(&oiP->txIpUdpHeaderList,
                                  &txIpUdpHeaderNtryP->listElement,
                                  &oiP->txIpUdpHeaderSpinLock);
  }

  NdisReinitializePacket(apacketP);
  if (KeGetCurrentIrql() == DISPATCH_LEVEL) {
    NdisDprFreePacket(apacketP);
  }
  else NdisFreePacket(apacketP);
  finishIrp(irpP, astatus, userBytesSent);
}

/*
   The NDIS wrapper calls this only on an asynchronous transfer, i.e., only
   if our receive handler's call to NdisTransferData() returned with a pending
   status.  In order to centralize completion handling, if our receive
   handler is lucky enough to call NdisTransferData() and not be told to wait,
   it calls this function as if it were the mighty wrapper.

   In fact, since this is where we recycle a receive buffer, we have to
   call this function every time we consume a receive buffer, even if we
   don't use NdisTransferData().  This happens in promiscuous mode.

   Since we allow a sloppy bootp daemon to send us broadcast packets,
   there's a possibility that this packet isn't really for us.  If we
   can tell that it isn't we bail out here.  Of course, this is
   germane only if we aren't in promiscuous mode.  By the way, the
   reason we do this checking here rather than in the receive handler
   is that it involves data that just might not be in the lookahead
   buffer.  We hate having to use the costly MmGetSystemAddressForMdl
   here, but the only alternative would be to let the application do the
   checking.
*/

VOID BgpsTransferDataCompleteHandler(IN NDIS_HANDLE abindingContext,
                                     IN PNDIS_PACKET apacketP,
                                     IN NDIS_STATUS astatus,
                                     IN UINT abytesTransferred)
{
  OpenInstance* oiP = (OpenInstance*)abindingContext;
  BgpsReserved* reservedP = (BgpsReserved*)(apacketP->ProtocolReserved);
  PIRP irpP = reservedP->irpP;

  BgpsTrace2Mask(DBG_RECEIVE,
                 "BgpsTransferDataCompleteHandler status = %x", astatus);

  if ((!oiP->promiscuous) && (astatus == NDIS_STATUS_SUCCESS)) {
    UCHAR* packetMACAddress =
      ((BootpBuf*)MmGetSystemAddressForMdl(irpP->MdlAddress))->chaddr;
    if (RtlCompareMemory(packetMACAddress, oiP->macAddress, MacAddressSize)
        != MacAddressSize) {
      KIRQL irql;
      BgpsTraceMask(DBG_RECEIVE, "Bootp response for wrong machine");
      IoAcquireCancelSpinLock(&irql);
      InsertTailList(&oiP->rcvList, &reservedP->listElement);
      IoSetCancelRoutine(reservedP->irpP,
                         BgpsCancelReadIrp); // Cancelable again!
      IoReleaseCancelSpinLock(irql);
      return;
    }
  }
  NdisReinitializePacket(apacketP); // recycle the packet
  if (KeGetCurrentIrql() == DISPATCH_LEVEL) {
    NdisDprFreePacket(apacketP);
  }
  else NdisFreePacket(apacketP); // Put the packet on the free queue
  if (astatus != STATUS_CANCELLED) finishIrp(irpP, astatus, abytesTransferred);
}

VOID BgpsResetCompleteHandler(IN NDIS_HANDLE abindingContext,
                              IN NDIS_STATUS astatus)
{
  BgpsTrace("ResetCompleteHandler Entry...");
}

VOID BgpsRequestCompleteHandler(IN NDIS_HANDLE abindingContext,
                                IN PNDIS_REQUEST andisRequest,
                                IN NDIS_STATUS astatus)
{
  POpenInstance oiP = (POpenInstance)abindingContext;
  PBgpsReqNtry reqNtryP = CONTAINING_RECORD(andisRequest, BgpsReqNtry,
                                            request);
  PIRP irpP = reqNtryP->irpP;

  BgpsTrace2("RequestCompleteHandler status = %x", astatus);

  NdisInterlockedInsertTailList(&oiP->requestList, &reqNtryP->listElement,
                                &oiP->requestSpinLock);
  if (irpP) {
    finishIrp(irpP, astatus, 0);
    if (irpP == oiP->openOrCloseIrpP) { // Completing MJ_CREATE
      if (NT_SUCCESS(astatus)) { // Good!
        UINT i;
        oiP->openOrCloseIrpP = 0;
        for (i = 0; i < MAX_TRANSMITS; i++) {
          if (oiP->u1.medium == NdisMedium802_3) {
            RtlCopyBytes(&oiP->txMacHeaders[i].payload.eth,
                         &ethernetHeaderStencil, sizeof(RFC894Header));
            RtlCopyBytes(&oiP->txMacHeaders[i].payload.eth.source,
                         oiP->macAddress, MacAddressSize);
          }
          else {
            RtlCopyBytes(&oiP->txMacHeaders[i].payload.tr,
                         &tokenRingHeaderStencil, sizeof(TokenRingHeader));
            oiP->txMacHeaders[i].payload.tr.source[0] =
              (UCHAR)(oiP->macAddress[0] | TokenRingRoutingInfoIndicator);
            RtlCopyBytes(oiP->txMacHeaders[i].payload.tr.source + 1,
                         oiP->macAddress, MacAddressSize - 1);
          }
        }
      }
      else {                    // Bad; won't need OpenInstance stuff
        emptyOpenInstance(oiP);
        ExFreePool(oiP);
      }
    }
  }
}

PNDIS_PACKET _stdcall getRcvPacket(OpenInstance* aoiP)
{
  PNDIS_PACKET packetP = NULL;
  KIRQL irql;
  PLIST_ENTRY listEntryP;
  PBgpsReserved reservedP;

  IoAcquireCancelSpinLock(&irql);
  if (!IsListEmpty(&aoiP->rcvList)) {
    listEntryP = RemoveHeadList(&aoiP->rcvList);
    reservedP = CONTAINING_RECORD(listEntryP, BgpsReserved, listElement);
    IoSetCancelRoutine(reservedP->irpP, NULL); // Can't cancel it now.
    packetP  =  CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
  }
  IoReleaseCancelSpinLock(irql);
  
  return packetP;
}

/*
  Here we handle the packet we may be waiting for.

  First we check to see if the app's really waiting, i.e. if there's an
  event.  Then we make sure it's a recognizable bootp response.  If so,
  we move the data by NdisTransferData (this is necessary because the
  packet may be longer than the lookahead buffer.)  The processing is
  finished by the TransferDataComplete handler, which we call ourselves
  if the data transfer was finished by the time we got control back from
  NdisTransferData.
*/

UCHAR llcSignature[3] = { 0xaa, 0xaa, 3 };

NDIS_STATUS BgpsReceiveHandler(IN NDIS_HANDLE abindingContext,
                               IN NDIS_HANDLE amacReceiveContext,
                               IN PVOID aheaderBuffer,
                               IN UINT aheaderBufferSize,
                               IN PVOID alookaheadBuffer,
                               IN UINT alookaheadBufferSize,
                               IN UINT apacketSize)
{
  USHORT packetType = 0xffff;   // initialized to non-IP value
  UINT ipOffset = 0;            // LLC/SNAP will augment this.
  OpenInstance* oiP = (OpenInstance*)abindingContext;

  BgpsTrace2Mask(DBG_RX_ALL, "ReceiveHandler packet size = %d", apacketSize);

  if (RtlCompareMemory(alookaheadBuffer, llcSignature, sizeof(llcSignature))
      == sizeof(llcSignature)) {
    LLCSNAP* llcSnapP = ((LLCSNAP*)alookaheadBuffer);
    packetType = llcSnapP->packetType;
    ipOffset = sizeof(LLCSNAP);
  }
  else if (aheaderBufferSize == sizeof(RFC894Header)) {
    packetType = ((RFC894Header*)aheaderBuffer)->packetType;
  }
  else if (!oiP->promiscuous) return NDIS_STATUS_NOT_ACCEPTED;

  if (oiP->promiscuous) {
    /*
      Poor man's packet sniffer: If packet meets our criteria, dump header
      length, header and packet "data".  If it's token ring we only compare
      on the last 5 bytes of the MAC address because of TR's habit of
      setting the most significant bit.
    */
    UINT dataRoom = sizeof(BillgpBuf) - aheaderBufferSize;
    UINT dataSize =
      (alookaheadBufferSize <= dataRoom) ? alookaheadBufferSize : dataRoom;
    UINT sourceOffset = (oiP->u1.medium == NdisMedium802_3) ?
      MacAddressSize : MacAddressSize + 3;
    UCHAR* ourAddress =  (oiP->u1.medium == NdisMedium802_3) ?
      oiP->macAddress : oiP->macAddress + 1;
    UINT comparisonLen = (oiP->u1.medium == NdisMedium802_3) ?
      MacAddressSize : MacAddressSize - 1;
    if ((apacketSize < 328) ||   // too small for bootp
        (apacketSize > 360) ||   // almost certainly too big
        (RtlCompareMemory(((UCHAR*)aheaderBuffer) + sourceOffset, // our query
                          ourAddress, comparisonLen) == comparisonLen)) {
      return NDIS_STATUS_NOT_ACCEPTED;
    }
    else {
      PNDIS_PACKET packetP = getRcvPacket(oiP);
      UINT totalBufferLength;
      UINT bytesTransferred = 0;
      if (packetP == NULL) return NDIS_STATUS_NOT_ACCEPTED;
#pragma warning(disable: 4127 4702)
      NdisQueryPacket(packetP, NULL, NULL, NULL, &totalBufferLength);
#pragma warning(default: 4127 4702)
      if (totalBufferLength < apacketSize) {
        KIRQL irql;
        BgpsTrace3("Receive packet length = %d but buffer length = %d",
                   apacketSize, totalBufferLength);
        IoAcquireCancelSpinLock(&irql);
        InsertTailList(&oiP->rcvList, &RESERVED(packetP)->listElement);
        IoSetCancelRoutine(RESERVED(packetP)->irpP,
                           BgpsCancelReadIrp); // Cancelable again!
        IoReleaseCancelSpinLock(irql);
        return NDIS_STATUS_NOT_ACCEPTED;
      }
      NdisTransferData(&oiP->transferDataStatus,
                       oiP->adapterHandle,
                       amacReceiveContext, 0, apacketSize,
                       packetP, &bytesTransferred);
      if (oiP->transferDataStatus != NDIS_STATUS_PENDING) {
        BgpsTransferDataCompleteHandler(abindingContext,
                                        packetP,
                                        oiP->transferDataStatus,
                                        bytesTransferred);
      }
    }
  }
  else if (packetType == PacketTypeIP) {
    IPHeader* ipP = (IPHeader*)((UCHAR*)alookaheadBuffer + ipOffset);

    if (ipP->protocol == IPPROTO_UDP) {
      UINT ipHeaderLength = (ipP->versionAndHeaderLength & 0xf) * 4;
      UINT totalHeaderLength = ipOffset + ipHeaderLength + sizeof(UDPHeader);
      UDPHeader* udpP = (UDPHeader*)(((UCHAR*)ipP)+ipHeaderLength);

      if ((totalHeaderLength < alookaheadBufferSize) &&
          (totalHeaderLength < apacketSize) && 
          (udpP->destinationPort == NetOrderPortBootpClient)) {
        UINT dataSize = apacketSize - totalHeaderLength;
        UINT bytesTransferred = 0;
        PNDIS_PACKET packetP = getRcvPacket(oiP);
        UINT totalBufferLength;

        if (packetP == NULL) {
          BgpsTrace("Ran out of receive packets");
          return NDIS_STATUS_NOT_ACCEPTED;
        }

        BgpsTraceMask(DBG_RECEIVE, "Received on bootp client port");
#pragma warning(disable: 4127 4702)
        NdisQueryPacket(packetP, NULL, NULL, NULL, &totalBufferLength);
#pragma warning(default: 4127 4702)
        if (dataSize > totalBufferLength) dataSize = totalBufferLength;
        NdisTransferData(&oiP->transferDataStatus,
                         oiP->adapterHandle,
                         amacReceiveContext, totalHeaderLength, dataSize,
                         packetP,
                         &bytesTransferred);
        if (oiP->transferDataStatus != NDIS_STATUS_PENDING) {
          BgpsTransferDataCompleteHandler(abindingContext,
                                          packetP,
                                          oiP->transferDataStatus,
                                          bytesTransferred);
        }
      } else return NDIS_STATUS_NOT_ACCEPTED;
    } else return NDIS_STATUS_NOT_ACCEPTED;
  } else return NDIS_STATUS_NOT_ACCEPTED;

  return(NDIS_STATUS_SUCCESS);
}

VOID BgpsReceiveCompleteHandler(IN NDIS_HANDLE abindingContext)
{
}

VOID BgpsStatusHandler(IN NDIS_HANDLE abindingContext,
                       IN NDIS_STATUS ageneralStatus,
                       IN PVOID astatusBuffer,
                       IN UINT astatusBufferSize)
{
  BgpsTrace("StatusHandler Entry...");
}

VOID BgpsStatusCompleteHandler(IN NDIS_HANDLE abindingContext)
{
  BgpsTrace("StatusCompleteHandler Entry...");
}

/* This function is called with the Cancel spin lock because it's been
   assigned to the IRP using IoSetCancelRoutine.  We have to find this
   IRP's packet on the receive list, remove it, and then recycle it.

   Bug alert: This algorithm is from Art Baker's book.  Both _NT Insider_'s
   series on cancel operations and section 12.3 in the DDK Kernel Mode Driver
   Design Guide's algorithm for drivers without a StartIo routine recommend
   ignoring the input IRP and hitting the first queued IRP with cancel
   set to TRUE.
*/

void BgpsCancelReadIrp(IN PDEVICE_OBJECT adeviceObjectP, IN PIRP airpP)
{
  PIO_STACK_LOCATION ioslP = IoGetCurrentIrpStackLocation(airpP);
  POpenInstance oiP = ioslP->FileObject->FsContext;
  BOOLEAN foundIt = FALSE;
  PLIST_ENTRY nextLinkP;

  BgpsTrace("BgpsCancelReadIrp");

  for (nextLinkP = oiP->rcvList.Flink;
       nextLinkP != &oiP->rcvList;
       nextLinkP = nextLinkP->Flink) {
    PBgpsReserved reservedP =
      CONTAINING_RECORD(nextLinkP, BgpsReserved, listElement);
    PNDIS_PACKET packetP  =
      CONTAINING_RECORD(reservedP, NDIS_PACKET, ProtocolReserved);
    if (reservedP->irpP == airpP) {
      foundIt = TRUE;
      RemoveEntryList(nextLinkP);
      NdisReinitializePacket(packetP); // recycle the packet
      NdisFreePacket(packetP);  // Put the packet on the free queue
      break;
    }
  }

  IoReleaseCancelSpinLock(airpP->CancelIrql);
  if (foundIt) finishIrp(airpP, STATUS_CANCELLED, 0);
}

VOID
BgpsBindAdapterHandler(OUT PNDIS_STATUS astatus,
                       IN  NDIS_HANDLE abindAdapterContext,
                       IN  PNDIS_STRING anadapterName,
                       IN  PVOID asystemSpecific1,
                       IN  PVOID asystemSpecific2)
{
    BgpsTrace("BgpsBindAdapterHandler\n");
    *astatus = NDIS_STATUS_FAILURE;
}

VOID
BgpsUnbindAdapterHandler(OUT PNDIS_STATUS astatus,
                         IN  NDIS_HANDLE aprotocolBindingContext,
                         IN  NDIS_HANDLE anunbindAdapterContext)
{
    BgpsTrace("BgpsUnbindAdapterHandler\n");
    *astatus = NDIS_STATUS_FAILURE;
}
